==> hourglass-1.0.0.tar.gz: hourglass-1.0.0/src/work.c <==
/*
 *
 * Copyright (c) 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Hourglass.
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation is hereby granted without fee, provided that the
 * above copyright notice and this permission/disclaimer notice is
 * retained in all copies or modified versions, and that both notices
 * appear in supporting documentation.  THE COPYRIGHT HOLDERS PROVIDE
 * THIS SOFTWARE "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE COPYRIGHT
 * HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 * RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu
 * any improvements that they make and grant redistribution rights to the
 * University of Utah.
 * 
 * Author: John Regehr (regehr@cs.utah.edu)
 */

#include "hourglass.h"

static void record_gap (cycle_time my_gap) 
{
  if (my_gap/GAP_HIST_DIVISOR > NUM_GAP_HIST_BUCKETS) {
    gap_hist_toobig++;
  } else {
    gap_hist[my_gap/GAP_HIST_DIVISOR]++;
  }
}

static int get_new_rec (void)
{
  long n;

  n = my_InterlockedIncrement (&current_record);

  if (n >= NUM_TRACE_RECORDS) {
    printf ("oops: no trace records left - exiting\n");
    exit (-1);
  }

  return n;
}

// this means nothing
static int funky_func (int x)
{
  return ((-2*x) << 3) - 775;
}

static cycle_time poll (int id, unsigned int *pos)
{
  cycle_time now, this_gap, runtime;

  thrd_data[id].workdone++;

  if (pos) {
    thrd_data[id].scanmem[*pos] = funky_func (thrd_data[id].scanmem[*pos+1]);
    (*pos)++;
    if ((*pos)+1 == (thrd_data[id].memsize / sizeof (int))) {
      *pos = 0;
    }
  }

  now = rdtsc ();
  this_gap = now - thrd_data[id].last_exec;

  // negative this_gap indicates likely processor switch

  // if id == num_threads, we're profiling and don't want to record
  // gaps

  if (id != num_threads && (this_gap > gap || this_gap < 0)) {

    int rec = thrd_data[id].reserved_record;
    thrd_data[id].reserved_record = get_new_rec ();

    if (thrd_data[id].exec_start != 0) {
      trace_recs[rec].id = id;
      trace_recs[rec].start = 
	thrd_data[id].exec_start + polling_overhead;
      trace_recs[rec].end = 
	thrd_data[id].last_exec - polling_overhead;
    }

    thrd_data[id].exec_start = now;
    runtime = 2*polling_overhead;

  } else {
    runtime = this_gap;
    if (generate_gap_hist) {
      record_gap (this_gap);
    }
  }

  thrd_data[id].last_exec = now;
  
  return runtime;
}

static void run_for (int id, cycle_time amount, unsigned int *pos)
{
  cycle_time total = 0;

  do {
    total += poll (id, pos);
  } while (total < amount && thrd_data[id].status != STOP);
}

static void lat_test (int id)
{
  cycle_time lat_start = rdtsc ();

  while (thrd_data[id].status != STOP) {
    thrd_data[id].timers->timed_sleep (id, -thrd_data[id].app_period); 

    thrd_data[id].lat_slots[thrd_data[id].lat_slot] = rdtsc() - lat_start;
    thrd_data[id].lat_slot++;
    if (thrd_data[id].lat_slot == LAT_SLOTS) break;
  }
}

#define FUDGE 10

static void periodic_test (int id)
{
  cycle_time now, deadline;

  deadline = rdtsc();

  while (thrd_data[id].status != STOP) {

    run_for (id, us_to_cycle (thrd_data[id].app_amount), NULL);

    deadline += us_to_cycle (thrd_data[id].app_period);      
    now = rdtsc();

    if (now <= deadline) {
      thrd_data[id].hits++;
      if (thrd_data[id].work == PERIODIC) {
	thrd_data[id].timers->timed_sleep (id, mymin (0, cycle_to_us (now - deadline) + FUDGE));
      } else {
	deadline = now;
      }
    } else {
      thrd_data[id].misses++;
      deadline = now;
    }   
  }
}

static void cpu_yield_test (int id, int scan)
{
  unsigned int pos = 0;

  while (thrd_data[id].status != STOP) {
    run_for (id, us_to_cycle (thrd_data[id].app_amount), (scan) ? &pos : NULL);
    sched_yield ();
  }
}

static void cpu_test (int id, int scan)
{
  unsigned int pos = 0;

  while (thrd_data[id].status != STOP) {
    poll (id, (scan) ? &pos : NULL);
  }
}

void *monitor_thrd_func (void *arg)
{
  int id = (int)arg;

  assert (thrd_data[id].s_time != -1);

  set_priority (TIMER_PRI);

  usleep (INITIAL_DELAY);

  usleep ((unsigned int)(thrd_data[id].f_time));  

  if (thrd_data[id].status == GO) {
    thrd_data[id].status = STOP;
  }

  /*
   * now, sleep a long time -- do this instead of just exiting since
   * exiting might be slow -- sleeping monitor threads will get
   * cleaned up when main() exits 
   */
  sleep (1000000000);

  return NULL;
}

void *thrd_func (void *arg)
{
  int id = (int)arg;

  set_priority (thrd_data[id].pri);

  if (thrd_data[id].res_period != -1) {
    make_res (thrd_data[id].res_amount, 
	      thrd_data[id].res_period, 
	      thrd_data[id].res_soft, 
	      id);
  }

  usleep (INITIAL_DELAY);

  if (thrd_data[id].s_time != -1) {
    thrd_data[id].timers->timed_sleep (id, -thrd_data[id].s_time);  
  }

  thrd_data[id].workdone = 0;

  switch (thrd_data[id].work) {

  case LAT:
    lat_test (id);
    break;

  case PERIODIC:
  case CPU_PERIODIC:
    thrd_data[id].hits = thrd_data[id].misses = 0;
    periodic_test (id);
    break;
    
  case CPU:
    cpu_test (id, FALSE);
    break;
    
  case CPU_SCAN:
    if (thrd_data[id].memsize > 0) {
      cpu_test (id, TRUE);
    } else {
      cpu_test (id, FALSE);
    }
    break;

  case CPU_YIELD:
    cpu_yield_test (id, FALSE);
    break;
    
  case CPU_SCAN_YIELD:
    if (thrd_data[id].memsize > 0) {
      cpu_yield_test (id, TRUE);
    } else {
      cpu_yield_test (id, FALSE);
    }
    break;

  default:
    printf ("Hrm: thread %d has an unknown workload\n", id);
    exit (-1);
  }

  // assert (thrd_data[id].status == STOP);

  thrd_data[id].status = DONE;

  delete_res (id);

  return NULL;
}

==> hourglass-1.0.1b.tar.gz: hourglass-1.0.1b/src/work.c <==
/*
 *
 * Copyright (c) 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Hourglass.
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation is hereby granted without fee, provided that the
 * above copyright notice and this permission/disclaimer notice is
 * retained in all copies or modified versions, and that both notices
 * appear in supporting documentation.  THE COPYRIGHT HOLDERS PROVIDE
 * THIS SOFTWARE "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE COPYRIGHT
 * HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 * RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu
 * any improvements that they make and grant redistribution rights to the
 * University of Utah.
 * 
 * Author: John Regehr (regehr@cs.utah.edu)
 */

#include "hourglass.h"

static void record_gap (cycle_time my_gap) 
{
  if (my_gap/GAP_HIST_DIVISOR > NUM_GAP_HIST_BUCKETS) {
    gap_hist_toobig++;
  } else {
    gap_hist[my_gap/GAP_HIST_DIVISOR]++;
  }
}

static int get_new_rec (void)
{
  long n;

  n = my_InterlockedIncrement (&current_record);

  if (n >= NUM_TRACE_RECORDS) {
    printf ("oops: no trace records left - exiting\n");
    exit (-1);
  }

  return n;
}

// this means nothing
static int funky_func (int x)
{
  return ((-2*x) << 3) - 775;
}

static cycle_time poll (int id, unsigned int *pos)
{
  cycle_time now, this_gap, runtime;

  thrd_data[id].workdone++;

  if (pos) {
    thrd_data[id].scanmem[*pos] = funky_func (thrd_data[id].scanmem[*pos+1]);
    (*pos)++;
    if ((*pos)+1 == (thrd_data[id].memsize / sizeof (int))) {
      *pos = 0;
    }
  }

  now = rdtsc ();
  this_gap = now - thrd_data[id].last_exec;

  // negative this_gap indicates likely processor switch

  // if id == num_threads, we're profiling and don't want to record
  // gaps

  if (id != num_threads && (this_gap > gap || this_gap < 0)) {

    int rec = thrd_data[id].reserved_record;
    thrd_data[id].reserved_record = get_new_rec ();

    if (thrd_data[id].exec_start != 0) {
      trace_recs[rec].id = id;
      trace_recs[rec].start = thrd_data[id].exec_start + polling_overhead;
      trace_recs[rec].end   = thrd_data[id].last_exec  - polling_overhead;
      // rjd -- SMP aware gapping code
      /*if(rec == 0)  // special case
      {
          // first record, it must show no gap
          trace_recs[rec].gap = 0;
          thrd_data[id].first++;
      }
      else*/ if(thrd_data[id].first == 0)
      {
          //haven't made the first record for this thread, gap should be as
          //before
          trace_recs[rec].gap = 0;//(rec) ? trace_recs[rec].start - trace_recs[rec-1].end: 0;
          thrd_data[id].first++;
      }
      else
      {
        trace_recs[rec].gap   = thrd_data[id].next_gap;
      }
        thrd_data[id].next_gap = this_gap;
    }

    thrd_data[id].exec_start = now;
    runtime = 2*polling_overhead;

  } else {
    runtime = this_gap;
    if (generate_gap_hist) {
      record_gap (this_gap);
    }
  }

  thrd_data[id].last_exec = now;
  
  return runtime;
}

static void run_for (int id, cycle_time amount, unsigned int *pos)
{
  cycle_time total = 0;

  do {
    total += poll (id, pos);
  } while (total < amount && thrd_data[id].status != STOP);
}

static void lat_test (int id)
{
  cycle_time lat_start = rdtsc ();

  while (thrd_data[id].status != STOP) {
    thrd_data[id].timers->timed_sleep (id, -thrd_data[id].app_period); 

    thrd_data[id].lat_slots[thrd_data[id].lat_slot] = rdtsc() - lat_start;
    thrd_data[id].lat_slot++;
    if (thrd_data[id].lat_slot == LAT_SLOTS) break;
  }
}

#define FUDGE 10

static void periodic_test (int id)
{
  cycle_time now, deadline;

  deadline = rdtsc();

  while (thrd_data[id].status != STOP) {

    run_for (id, us_to_cycle (thrd_data[id].app_amount), NULL);

    deadline += us_to_cycle (thrd_data[id].app_period);      
    now = rdtsc();

    if (now <= deadline) {
      thrd_data[id].hits++;
      if (thrd_data[id].work == PERIODIC) {
	thrd_data[id].timers->timed_sleep (id, mymin (0, cycle_to_us (now - deadline) + FUDGE));
      } else {
	deadline = now;
      }
    } else {
      thrd_data[id].misses++;
      deadline = now;
    }   
  }
}

static void cpu_yield_test (int id, int scan)
{
  unsigned int pos = 0;

  while (thrd_data[id].status != STOP) {
    run_for (id, us_to_cycle (thrd_data[id].app_amount), (scan) ? &pos : NULL);
    sched_yield ();
  }
}

static void cpu_test (int id, int scan)
{
  unsigned int pos = 0;

  while (thrd_data[id].status != STOP) {
    poll (id, (scan) ? &pos : NULL);
  }
}

// rjd
static void io_dumb_test(int id)
{
    // allocate our data buffer
    thrd_data[id].dbuf = calloc(1, thrd_data[id].dsize);
    if(!thrd_data[id].dbuf)
    {
        perror("calloc");
        exit(-1);
    }

    // open the file
    if((thrd_data[id].fd = fopen(thrd_data[id].fname, "r+")) == NULL)
    {
        fprintf(stderr, "INSANITY: thread %d cannot open its file: ", id);
        perror("fopen");
        exit(-1);
    }
    
    // run the test
    while(thrd_data[id].status != STOP)
    {
        size_t b_read, b_write;
        
        // take a trace record
        poll(id, NULL);
        // read a record
        b_read = fread(thrd_data[id].dbuf, 1, thrd_data[id].dsize, 
                    thrd_data[id].fd);
        if(b_read < thrd_data[id].dsize)
        {
            if(!feof(thrd_data[id].fd))
            {
                fprintf(stderr, "thread %d: read %u bytes\n",
                            id, (unsigned int) b_read);
                perror("fread");
                exit(-1);
            }
            //rewind and try again
            fseek(thrd_data[id].fd, 0, SEEK_SET);
            continue;   // try again
        }
        fflush(thrd_data[id].fd);
       
        // overwrite that same record
        fseek(thrd_data[id].fd, 0-thrd_data[id].dsize, SEEK_CUR); 
       
        // take another trace record
        poll(id, NULL);
        
        // write it
        b_write = fwrite(thrd_data[id].dbuf, 1, b_read, thrd_data[id].fd);
        if(b_write < b_read)
        {
            fprintf(stderr, "thread %d: ", id);
            perror("fwrite");
            exit(-1);
        }
        fflush(thrd_data[id].fd);

        // workdone is the amount of data written plus 2 * the amount this loop
        // was run
        thrd_data[id].workdone += b_write;
    }

    free(thrd_data[id].fname);
    fclose(thrd_data[id].fd);
    free(thrd_data[id].dbuf);
}

void *monitor_thrd_func (void *arg)
{
  int id = (int)arg;

  assert (thrd_data[id].s_time != -1);

  set_priority (TIMER_PRI);

  usleep (INITIAL_DELAY);

  usleep ((unsigned int)(thrd_data[id].f_time));  

  if (thrd_data[id].status == GO) {
    thrd_data[id].status = STOP;
  }

  /*
   * now, sleep a long time -- do this instead of just exiting since
   * exiting might be slow -- sleeping monitor threads will get
   * cleaned up when main() exits 
   */
  sleep (1000000000);

  return NULL;
}

void *thrd_func (void *arg)
{
  int id = (int)arg;

  set_priority (thrd_data[id].pri);

  if (thrd_data[id].res_period != -1) {
    make_res (thrd_data[id].res_amount, 
	      thrd_data[id].res_period, 
	      thrd_data[id].res_soft, 
	      id);
  }

  usleep (INITIAL_DELAY);

  if (thrd_data[id].s_time != -1) {
    thrd_data[id].timers->timed_sleep (id, -thrd_data[id].s_time);  
  }

  thrd_data[id].workdone = 0;

  switch (thrd_data[id].work) {

  case LAT:
    lat_test (id);
    break;

  case PERIODIC:
  case CPU_PERIODIC:
    thrd_data[id].hits = thrd_data[id].misses = 0;
    periodic_test (id);
    break;
    
  case CPU:
    cpu_test (id, FALSE);
    break;
    
  case CPU_SCAN:
    if (thrd_data[id].memsize > 0) {
      cpu_test (id, TRUE);
    } else {
      cpu_test (id, FALSE);
    }
    break;

  case CPU_YIELD:
    cpu_yield_test (id, FALSE);
    break;
    
  case CPU_SCAN_YIELD:
    if (thrd_data[id].memsize > 0) {
      cpu_yield_test (id, TRUE);
    } else {
      cpu_yield_test (id, FALSE);
    }
    break;
  case IO_DUMB:
    io_dumb_test(id);
    break;

  default:
    printf ("Hrm: thread %d has an unknown workload\n", id);
    exit (-1);
  }

  // assert (thrd_data[id].status == STOP);

  thrd_data[id].status = DONE;

  delete_res (id);

  return NULL;
}
